# -*- coding: utf-8 -*-
"""
This Module contains all custom errors used in the package.
"""

from discord.ext.commands import CommandError


class NotMemberOfCorrectGuild(CommandError):
    pass


class IncorrectGuild(CommandError):
    pass


class Unauthorized(CommandError):
    pass


class InvalidOption(CommandError):
    pass

class UnbError(Exception):
    pass


class ResponseError(UnbError):
    pass


class UnbRateLimited(UnbError):
    pass



# -*- coding: utf-8 -*-

"""
This module holds functions for displaying different kinds of menus.
All menus are reaction based.
"""

import asyncio
from typing import List, Optional, Union

import discord
from discord import Embed
from discord.ext import commands


async def try_clear_reactions(msg):
    """helper function to remove reactions excepting forbidden
    either by context being a dm_channel or bot lacking perms"""

    if msg.guild:
        try:
            await msg.clear_reactions()
        except discord.errors.Forbidden:
            pass


async def arrows(ctx: commands.Context,
                 embed_list: List[Embed],
                 head: int = 0,
                 timeout: int = 30,
                 closed_embed: Optional[Embed] = None,
                 channel: Optional[discord.abc.Messageable] = None):
    """
    Sends multiple embeds with a reaction navigation menu.

    Args:
        ctx:
            The context where this function is called.
        embed_list  (List[Embed])
            An ordered list containing the embeds to be sent.
        head (int)
            The index in embed_list of the first Embed to be displayed.
        timeout (int: seconds)
            The time before the bot closes the menu.
            This is reset with each interaction.
        closed_embed (Optional[Embed]):
            The embed to be displayed when the user closes the menu.
            Defaults to plain embed with "Closed by user" in description

        channel (discord.abc.Messageable)
            The channel to be used for displaying the menu, defaults to ctx.channel.
    Returns:
        None
    """
    channel = channel or ctx.channel
    closed_embed = closed_embed or Embed(description="Closed by user")

    if len(embed_list) == 1:
        return await ctx.send(embed=embed_list[0])

    emojis = ['⏮', '◀', '▶', '⏭', '❌', '⏸']

    msg = await channel.send(embed=embed_list[head])

    for emoji in emojis:
        await msg.add_reaction(emoji)

    def check(payload):
        return all([
            payload.user_id == ctx.author.id,
            payload.emoji.name in emojis,
            msg.id == payload.message_id
        ])

    def return_head(head: int, emoji: str) -> Union[bool, int]:
        if emoji == '⏮':  # return to the first Embed
            return 0
        elif emoji == '◀':  # one left
            return head - 1 if head > 0 else 0
        elif emoji == '▶':  # one right
            return head + 1 if head < (len(embed_list) - 1) else len(embed_list) - 1
        elif emoji == '⏭':  # go to the end
            return len(embed_list) - 1
        elif emoji == '❌':  # close the menu
            return False
        elif emoji == '⏸':  # remove the reactions and keep the selected embed open.
            return True

    while True:
        try:
            payload = await ctx.bot.wait_for('raw_reaction_add', timeout=timeout, check=check)
        except asyncio.TimeoutError:
            return await try_clear_reactions(msg)
        else:
            head = return_head(head, payload.emoji.name)
            if head is True:
                return await try_clear_reactions(msg)

            if head is False:
                await try_clear_reactions(msg)
                return await msg.edit(embed=closed_embed)

            elif isinstance(head, int):
                try:
                    await msg.remove_reaction(payload.emoji, ctx.author)
                except discord.errors.Forbidden:
                    pass

                await msg.edit(embed=embed_list[head])


async def confirm(ctx: commands.Context,
                  msg: discord.Message,
                  lock: Union[discord.Member, discord.Role, bool, None] = True,
                  timeout: int = 30) -> Optional[bool]:
    """Helps to create a reaction menu to confirm an action.
    :Parameters:
        ctx                    - the context for the menu
        msg                    - the message to confirm or deny by the user.
        lock                   -
            :Bool:             -
                True (default) - the menu will only listen for the author's reactions.
                False          - ANY user can react to the menu
            :discord.Member:   - only target member will be able to react
            :discord.Role:     - ANY user with target role will be able to react.
        timeout (seconds)      - Timeout before the menu closes.

    Returns:
        True if the message was confirmed
        False if it was denied
        None if timeout
    """

    emojis = ['👍', '❌']
    for emoji in emojis:
        await msg.add_reaction(emoji)

    def check(payload):
        _checks = [
            payload.user_id != ctx.bot.user.id,
            payload.emoji.name in emojis,
            payload.message_id == msg.id,
        ]
        if lock:
            if isinstance(lock, bool):
                _checks.append(payload.user_id == ctx.author.id)
            elif isinstance(lock, discord.Member):
                _checks.append(payload.user_id == lock.id)
            elif isinstance(lock, discord.Role):
                _checks.append(lock in ctx.guild.get_member(payload.user_id).roles)
        return all(_checks)

    try:
        payload = await ctx.bot.wait_for('raw_reaction_add', check=check, timeout=timeout)
    except asyncio.TimeoutError:
        await try_clear_reactions(msg)
        return None
    else:
        await try_clear_reactions(msg)
        if payload.emoji.name == '👍':
            return True
        else:
            return False



# -*- coding: utf-8 -*-
"""
This module holds parsers and converters to use with user input
Functions here can be used as type hints which discord.py will use as
custom converters or they can be used as regular functions.
Example:
    ```
    @bot.command()
    async def timedelta(ctx, time: dpytools.parsers.parse_time):
        await ctx.send(f"time delta is: {time}")
    ```
    above command will be called like this: `!timedelta 2h30m`
    and it will send a message with "time delta is: 2:30:00"

    This way you dont have to manually parse the time string to a timedelta object.
    the parameter "time" will be of type timedelta.
"""
import re
from collections import ChainMap
from datetime import timedelta
from typing import Dict


class InvalidTimeString(Exception):
    pass


def to_upper(string: str) -> str:
    """
    Converts :string: to upper case. Intended to be used as argument converter.
    Args:
        string: string to format

    Returns:
        string to upper case
    """
    return string.upper()


def to_lower(string: str) -> str:
    """
    Converts :string: to lower case. Intended to be used as argument converter.
    Args:
        string: string to format

    Returns:
        string to lower case
    """
    return string.lower()


def time_parser(string: str) -> timedelta:
    """
    Converts a string with format <number>[s|m|h|d] to a timedelta object
    <number> must be convertible to float.
    Uses regex to match groups and consumes all groups into one timedelta.

    Units:
        s: second
        m: minute
        h: hour
        d: day
        w: weeks

    Args:
        string: with format <number>[s|m|h|d].
            parse_time("2h") == timedelta(hours=2)

    Returns:
        timedelta

    Raises:
        ValueError: if string number cannot be converted to float
        InvalidTimeString: if string isn't in the valid form.

    """
    units = {
        's': 'seconds',
        'm': 'minutes',
        'h': 'hours',
        'd': 'days',
        'w': 'weeks'
    }

    time_pattern = r"(\d+\.?\d?[s|m|h|d|w]{1})\s?"

    def parse(time_string: str) -> Dict:

        unit = time_string[-1]

        amount = float(time_string[:-1])
        return {units[unit]: amount}

    if matched := re.findall(time_pattern, string, flags=re.I):
        time_dict = dict(ChainMap(*[parse(d) for d in matched]))
        return timedelta(**time_dict)
    else:
        raise InvalidTimeString("Invalid string format. Time must be in the form <number>[s|m|h|d|w].")



# -*- coding: utf-8 -*-
"""
This module holds functions to work with embeds in different ways.
"""

import discord
from discord import Embed
from discord.ext.commands import Paginator
from typing import List, Optional, Union


def paginate_to_embeds(title: str,
                       description: str,
                       max_size: int = 2000,
                       prefix: Optional[str] = "",
                       suffix: Optional[str] = "",
                       color: Union[discord.Color, int, None] = None
                       ) -> List[Embed]:
    """
    Facilitates sequential embed menus.
    Returned embeds share title, have description split at :max_length: and are added index/pages at footer to keep
    track of pages.

    Args:
        title: Shared by all embeds
        description: String to be split at :max_lenght: per embed.
        max_size: Maximum amount of characters per embed. Discord's limit is 2000.
        prefix: Defaults to "" it will be appended at the start of the description of each embed.
                Useful for codeblocks (use triple back quotes).
        suffix: Same as :prefix: but at the end of the text.
        color: color to use for the embed. Accepts int (decimal or hex) or discord.Color/discord.Colour.

    Returns:
        List[Embed]
    """
    embeds = []
    to_list = description.split("\n")
    paginator = Paginator(prefix=prefix, suffix=suffix, max_size=max_size)
    for line in to_list:
        paginator.add_line(line)
    for i, page in enumerate(paginator.pages):
        embeds.append(Embed(title=title,
                            description=page,
                            color=color, )
                      .set_footer(text=f"page: {i + 1}/{len(paginator.pages)}"))
    return embeds



# -*- coding: utf-8 -*-
"""
Module that holds methods related to discord's bot client method `wait_for`
They're useful for retrieving information from users.
This functions and classes expect at least discord.Context
so they're have to be use with the commands extension.

Example:
     ```
        @bot.command()
        async def test(ctx):
            msg_a = await ctx.send('reply with a message containing only alpha, beta or gamma')
            msg_b = await wait_for_regex(ctx,
                                         pattern=r"^(alpha|beta|gamma)$",
                                         timeout=15)
            out = "got a match" if msg_b else "I didn't get a match within 15 seconds. Try again"
            await msg_a.edit(content=out)
     ```
"""

import asyncio
import re
from typing import Union, Optional

import discord
from discord.ext import commands


class BaseLock:
    """
    Callable class to use with the :check: parameter in client.wait_for('message')
    The default behavior only requires to pass the context in the constructor and will perform these checks:
        `ctx.author == message.author and ctx.channel == message.channel`
    If channel is passed then it will check `channel == message.channel`
    If lock parameter is passed (differently than True) then it'll check the object against message.author.

    The constructor Raises ValueError if channel is not a GuildChannel when lock is type Role.
    """

    def __init__(self,
                 ctx: commands.Context,
                 channel: Union[discord.TextChannel,
                                discord.abc.PrivateChannel] = None,
                 lock: Union[discord.Role,
                             discord.User,
                             discord.Member,
                             bool] = True,
                 ):
        self.ctx = ctx
        self.channel = channel or ctx.channel
        if isinstance(lock, discord.Role) and not isinstance(self.channel, discord.TextChannel):
            raise ValueError(f":lock: parameter is a role but the target channel is not a guild TextChannel")
        self.lock = lock

    def __call__(self, message: discord.Message) -> bool:
        checks = [message.channel == self.channel]

        if self.lock is True:
            checks.append(self.ctx.author == message.author)
        elif isinstance(self.lock, discord.Role):
            checks.append(self.lock in self.ctx.author.roles)
        elif isinstance(self.lock, (discord.Member, discord.User)):
            checks.append(self.lock == self.ctx.author)

        return all(checks)


async def wait_for_regex(ctx: commands.Context,
                         pattern: str,
                         ignore_case: bool = False,
                         timeout: int = 30,
                         channel: Union[discord.TextChannel,
                                        discord.abc.PrivateChannel] = None,
                         lock: Union[discord.Role,
                                     discord.Member,
                                     bool] = True,
                         ) -> Optional[discord.Message]:
    """
    Waits for a message that contains a match for the passed :pattern:

    Args:
        ctx: the command context
        pattern: regex string to look in the message
        ignore_case: Defaults to False. If True sets re.I as flag.
        timeout: time in seconds to wait for the appropriate message
        channel: The channel where the message should come from. Defaults to ctx.channel.
        lock:
            discord.Role -> Only members with this role can answer
            discord.Member -> only specified member can answer
            bool: True -> Only ctx.author can answer
            bool: False -> Anyone can answer as long as its in the correct channel within the timeout window
    Returns:
        Optional[discord.Message]: returns None if timeout is specified and reached
        and the rest of the checks arent passed or no match for the pattern is found

    """

    class Check(BaseLock):
        def __call__(self, msg):
            if super().__call__(msg):
                return re.match(pattern, msg.content, flags=re.I if ignore_case else 0)

    check = Check(ctx, channel, lock)

    try:
        message = await ctx.bot.wait_for('message', check=check, timeout=timeout)
    except asyncio.TimeoutError:
        return
    else:
        return message


async def wait_for_author(ctx: commands.Context,
                          stop: str = 'cancel',
                          timeout: Optional[int] = 30,
                          ) -> Optional[discord.Message]:
    """
    This function returns a single message from ctx.author in ctx.channel.
    Args:
        ctx: the command context
        stop: string to stop this function
        timeout: time in seconds to wait for a reply or None if the function should wait forever (usual maximum is a day)

    Returns:
        None: if :timeout: is reached or if :stop: string is passed
        discord.Message: User's reply message.
    """
    try:
        message = await ctx.bot.wait_for('message', timeout=timeout, check=BaseLock(ctx))
    except asyncio.TimeoutError:
        return
    else:
        if message.content.lower() == stop:
            return
        else:
            return message



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dpytools']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'dpytools',
    'version': '0.0.17a0',
    'description': 'Simple tools to build discord bots using discord.py',
    'long_description': "\n[![PyPI status](https://img.shields.io/pypi/status/dpytools.svg)](https://pypi.python.org/pypi/dpytools/)\n[![PyPI version fury.io](https://badge.fury.io/py/dpytools.svg)](https://pypi.python.org/pypi/dpytools/)\n[![PyPI license](https://img.shields.io/pypi/l/dpytools.svg)](https://pypi.python.org/pypi/dpytools/)\n\n\n# dpytools\nToolset to speed up developing discord bots using discord.py\n\n<hr>\n\n## Status of the project\n\nEarly development. As such it's expected to be unstable and unsuited for production.\n\nAll the presented tools have stringdocs.\n<hr>\n\n# Instalation\nInstall the latest version of the library with pip.\n```\npip install -U dpytools\n```\n\n# Contributing\nFeel free to make a pull request.",
    'author': 'chrisdewa',
    'author_email': 'alexdewa@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/chrisdewa/dpytools',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
Common general purpose commands

Add them to your bot by simply doing
    ```
        bot.add_command(command)
    ```
"""

import os
from typing import Optional

from discord import Embed, Color
from discord.ext import commands
from discord.ext.commands import ExtensionError


@commands.command(aliases=['lat'])
async def latency(ctx: commands.Context):
    """
    Command to disply latency in ms.
    Latency is displayed on plain embed in the description field
    Args:
        ctx: discord.Context

    Returns:
        None
    """
    await ctx.send(embed=Embed(description=f"Latency: {round(ctx.bot.latency * 1000, 1)}ms"))


@commands.command(hidden=True)
@commands.is_owner()
async def cogs(ctx: commands.Context,
               option: Optional[str],
               cog: Optional[str],
               cogs_dir: str = "./cogs"):
    """
    Command to load, unload and reload extensions.

    Checks:
        commands.is_owner

    Attributes:
        hidden=True

    Args:
        ctx: commands.Context
        option:
            load: loads :cog:
            unload: unloads :cog:
            reload: relaods :cog:
            list: lists available cogs in :cogs_dir:

        cog: filename of the extension to load, without the file extension ('my_cog' not 'my_cog.py')
        cogs_dir: directory to operate command, defaults to ./cogs


    """
    if option.lower() not in ['load', 'unload', 'reload', 'list']:
        await ctx.send(embed=Embed(
            description=f"{option} is not valid. Valid options are:'load', 'unload', 'reload' and 'list' "))

    bot = ctx.bot

    actions = {
        'load': bot.load_extension,
        'unload': bot.unload_extension,
        'reload': bot.reload_extension
    }

    cogs = [found_cog for dp, dn, fn in
            os.walk(os.path.expanduser(cogs_dir)) for f in fn
            if '__pycache__' not in (found_cog := os.path.join(dp, f)
                                     .replace(cogs_dir, '')
                                     .replace('\\', '.')
                                     .replace('/', '.')[1:-3])]

    if option.lower() == 'list':
        await ctx.send(embed=Embed(
            title="Available cogs:",
            description="".join(f"{i + 1}) {cog}\n" for i, cog in enumerate(cogs)),
            color=Color.blue()
        ))
    else:
        action = actions[option]

        if cog.lower() == 'all':
            msg = await ctx.send(embed=Embed(description=f'Performing: {action.__name__} on all extensions'))
            for cog in cogs:
                try:
                    action(f"{cogs_dir[2:]}.{cog}")
                except ExtensionError as E:
                    await msg.edit(embed=Embed(description=f"Action could not be completed.\n"
                                                           f"Reason:\n"
                                                           f"```{E}```"))
                else:
                    await msg.edit(embed=Embed(description=f'**Done!**'))
        elif cog not in cogs:
            await ctx.send(embed=Embed(
                description=f"{cog} is not a valid cog, check spelling and verify directory."))
        else:
            msg = await ctx.send(embed=Embed(description=f'Performing: {action.__name__} on {cog}'))
            try:
                action(f"{cogs_dir[2:]}.{cog}")
            except ExtensionError as E:
                await msg.edit(embed=Embed(description=f"Action could not be completed.\n"
                                                       f"Reason:\n"
                                                       f"```{E}```"))
            else:
                await msg.edit(embed=Embed(description=f'**Done!**'))



# -*- coding: utf-8 -*-
"""
This module holds different command checks to add as decorator for custom commands.
Use this checks on your code like this:
    ```
    @bot.command()
    @admin_or_roles('Moderator')
    async def mute(ctx, member: discord.Member):
        ...
    ```
Or inside Cogs:
    ```
    @commands.command()
    @only_this_guild(123456789...)
    async def magic(self, ctx):
        ...
    ```

"""


from discord.ext import commands
from discord.ext.commands import PrivateMessageOnly
from typing import Union, List
from discord.utils import get

from dpytools.errors import Unauthorized, IncorrectGuild, NotMemberOfCorrectGuild


def admin_or_roles(*roles: Union[int, str]) -> commands.check:
    """
    This check returns true under these conditions:
        The command is run in a guild AND:
            ctx.author has admin permissions OR has any role in :roles:

    Raises:
        commands.NoPrivateMessage if ran from DM
        Unauthorized if user doesn't have correct roles or admin permissions
        ValueError if value passed on role_ids is not int or str
    """

    async def predicate(ctx):
        if ctx.guild is None:
            raise commands.NoPrivateMessage()

        if ctx.author.guild_permissions.administrator is True:
            return True

        discord_roles = []
        for role in roles:
            if isinstance(role, str):
                discord_roles.append(get(ctx.guild.roles, name=role))
            elif isinstance(role, int):
                discord_roles.append(ctx.guild.get_role(role))
            else:
                raise ValueError(f"int or str was expected but received {type(role)}")
        if any([role for role in discord_roles if role in ctx.author.roles]):
            return True
        else:
            raise Unauthorized("User doesn't have admin permissions or specified roles")

    return commands.check(predicate)


def only_this_guild(guild_id: int) -> commands.check:
    """This check returns true if ctx.guild has same id as param :guild_id:

    Raises:
        commands.NoPrivateMessage if ran from DM
        Custom IncorrectGuild commands.CommandError if id doesn't check

    """

    async def predicate(ctx):
        if ctx.guild is None:
            raise commands.NoPrivateMessage("Command was called from a direct message.")
        elif guild_id != ctx.guild.id:
            raise IncorrectGuild(f"Command was called from a guild with id different than specified in check")
        return True

    return commands.check(predicate)


def dm_from_this_guild(guild_id: int, delete: bool = False) -> commands.check:
    """
    The check returns true only if the specified :guild_id: is found within the bot's guilds.
    For this check to work the guilds and members intents must be enabled.

    Args:
        guild_id: id of the guild the author of the context must be a part of for check to return True
        delete: if True, it will try to delete ctx.message. This will only happen if the message is called
                from the specificed guild_id, if the command is called from another guild the check will
                just return False.

            Possible exceptions by discord.py library:
                commands.Forbidden – You do not have proper permissions to delete the message.
                commands.NotFound – The message was deleted already
                commands.HTTPException – Deleting the message failed.

    Returns:
        commands.check

    Raises:
        commands.PrivateMessageOnly: if called from a guild
        dpytools.errors.NotMemberOfCorrectGuild: if not a member of the specified guild
    """
    async def predicate(ctx):
        if ctx.guild and ctx.guild.id == guild_id:
            if delete is True:
                await ctx.message.delete()
            raise PrivateMessageOnly("Command was called from the guild instead of a direct message.")
        elif ctx.guild:
            return False

        guild = ctx.bot.get_guild(guild_id)

        if ctx.author in guild.members:
            return True
        else:
            raise NotMemberOfCorrectGuild("This command is unavailable to you.")

    return commands.check(predicate)

